/*
 * upc_tester.c -- active testing for upc
 *
 */

#ifndef UPCR_THRILLE_TESTER
#define UPCR_THRILLE_TESTER  // Make sure all recursive calls into thrille are properly suffixed
#endif
#include "upct_types.h"
#include "utils/str_int_map.h"


#define UPCT_RACE_ID_ENV_STRING	"UPCT_RACE_ID"
#define UPCT_RACE_RESOLVE_ENV_STRING	"UPCT_RACE_RESOLVE_DIR"

struct _race_target_t {
    const char* sig;
};

typedef struct {
    addr_interval mem;
    int thread;
    int phase;
} tester_info_t;

//typedef tester_info_t tester_pending_t[2];

/* Shared Data */
typedef bupc_sem_t* ptr_sem;            // To enforce thread affinity
shared [1] ptr_sem semaphores[THREADS]; // otherwise, runtime assertion failure
typedef upc_lock_t* ptr_lock;
shared [1] ptr_lock locks[THREADS];
shared [2] tester_info_t pending[2*THREADS];
shared [2] tester_info_t ot_races[2*THREADS];
shared int num_thrash = 0;
shared int num_reproduced = 0;

shared double gprob = 0.75;


/* Per Thread Data */
//upc_lock_t *thrille_lock;
thrille_info_t thrille_info = { 0, 0, 0 };
race_info_t target[2];
int race_resolve_dir;
double lprob = 0.75;
tester_info_t *my_pending;
tester_info_t *my_ot_race;
tester_info_t race_event;
//race_info_t raceEvent;
bupc_tick_t timeout;
memory_pool persistent_mempool;
str_int_map* sig_map;
int sigs_found;

#define THRILLE_TEST_SINGLE     1       // Implies owned = N, access = WRITE
#define THRILLE_TEST_DOUBLE     2

int thrille_test_mode;

void thrille_init(void) {
    if (MYTHREAD == 0) {
        fprintf(stderr, "UPC Active Testing " THRILLE_VERSION " started (BACK_OFF=%g, EPS=%g)...\n", BACK_OFF, EPS);
    }

    char* race_id_str = bupc_getenv(UPCT_RACE_ID_ENV_STRING);
    int id = race_id_str ? atoi(race_id_str) : 1;
    char* race_resolve_str = bupc_getenv(UPCT_RACE_RESOLVE_ENV_STRING);
    race_resolve_dir = race_resolve_str ? atoi(race_resolve_str) : 0;
    if (MYTHREAD == 0)
        fprintf(stderr, "Reading race pair %d\n", id);

    read_race_pair(target, target + 1, id);
    thrille_test_mode = ((target[0].owned == target[1].owned) &&
            (target[0].access == target[1].access) &&
            (strcmp(target[0].sig, target[1].sig) == 0)) ? 
            THRILLE_TEST_SINGLE : THRILLE_TEST_DOUBLE;

    if (MYTHREAD == 0) {
        print_race_info_ns(target);
        fprintf(stderr, "\n===\n");
        print_race_info_ns(target + 1);
        fprintf(stderr, "\n");
        fprintf(stderr, "Mode: %s, Resolve order: %d\n", 
                thrille_test_mode == THRILLE_TEST_SINGLE ? "SINGLE" : "DOUBLE", 
                race_resolve_dir);
    }

    //setRaceTargets();

    //thrille_lock = upc_all_lock_alloc();
    locks[MYTHREAD] = upc_global_lock_alloc();
    //semaphores = (shared ptr_sem*) upc_all_alloc(THREADS, sizeof (ptr_sem));
    //semaphores[MYTHREAD] = bupc_sem_alloc(BUPC_SEM_BOOLEAN | BUPC_SEM_MPRODUCER
    //        | BUPC_SEM_SCONSUMER);

    //races = (shared race_info_t*) upc_all_alloc(THREADS, sizeof (race_info_t));
    //my_pending = (tester_pending_t*) &pending[MYTHREAD];
    //my_pending = (tester_info_t*) &pending[2 * MYTHREAD];
    my_ot_race = (tester_info_t*) &ot_races[2 * MYTHREAD];
    //my_pending[0].mem.start = 0;
    //my_pending[1].mem.start = 0;
    pending[2*MYTHREAD].mem.start = 0;
    pending[2*MYTHREAD + 1].mem.start = 0;

    thrille_info.inside_inst = 0;
    thrille_info.phase = 0;
    double ticks_per_us = 1000000.0 / bupc_ticks_to_ns(1000);
    timeout = SEM_WAIT_USEC * MAX_SEM_TRIES * ticks_per_us;
    
    new_mempool(&persistent_mempool);
    sig_map = str_int_map_new(&persistent_mempool, 1023);
    sigs_found = 0;

    //fprintf(stderr, "[%d] Timeout = %lld ticks\n", MYTHREAD, timeout);
    upc_barrier;

    thrille_info.initialized = 1;
}

void thrille_exit(int code)
{
}

/* Returns matching target (1 or 2) or 0 if no match
 */
int isRacingPair(const char* sig, int owned, int access) {
    for (int i = 0; i < thrille_test_mode; ++i) {
        if ((target[i].sig == sig) && (owned == target[i].owned) && 
                (access == target[i].access)) 
            return i + 1;
    } 
    
    int v = str_int_map_get(sig_map, sig, 0);
    if (v == 0) {
        int ret = 0;
        for (int i = 0; i < thrille_test_mode; ++i) {
            if (strcmp(sig, target[i].sig) == 0) {
                target[i].sig = sig;
                v |= (i + 1);
                if ((owned == target[i].owned) && (access == target[i].access))
                    ret = i + 1;
            }
        }
        str_int_map_put(sig_map, sig, v ? v : -1);
        return ret;
    }
    
    return 0;
    
#if 0
    //fprintf(stderr, "%s %s\n", sig, target1.sig);
    int isRacingPair1 = (owned == target1.owned)
            && (strcmp(sig, target1.sig) == 0)
            //&& (thrille_info.phase == target1.phase)
            ;

    int isRacingPair2 = (owned == target2.owned)
            && (strcmp(sig, target2.sig) == 0)
            //&& (thrille_info.phase == target2.phase)
            ;

    return isRacingPair1 + 2 * isRacingPair2;
#endif
}

int checkPending(int owner, int stmt) {
    upc_memget(my_ot_race, &pending[2 * owner], 2 * sizeof(tester_info_t));
    if (thrille_test_mode == THRILLE_TEST_SINGLE) {
        if (my_ot_race[0].mem.start) {
            return overlaps_interval(my_ot_race[0].mem, race_event.mem) ? 
                    my_ot_race[0].thread : BAD_RACE;
        }
    }
    else {
        if (my_ot_race[2-stmt].mem.start && 
                overlaps_interval(my_ot_race[2-stmt].mem, race_event.mem)) {
            return my_ot_race[2-stmt].thread;
        }
        if (my_ot_race[stmt-1].mem.start) {
            return BAD_RACE;
        }
    }
    return NO_RACE;
#if 0
    if (ot_race->addr[0]) {
        int sameAddr = (ot_race->addr[0] < raceEvent.addr[1] && 
                raceEvent.addr[0] < ot_race->addr[1]);
        int oneWrite = (raceEvent.access == WRITE ||
                ot_race->access == WRITE);
        int sameTarget = (raceEvent.race_reported == 1 && ot_race->race_reported == 1)
                || (raceEvent.race_reported == 2 && ot_race->race_reported == 2);
        if (sameAddr) {
            if (oneWrite && !sameTarget)
                //if (sameAddr && oneWrite && !sameTarget)
                return t;
            else
                return BAD_RACE;
        } else if (sameTarget)
            return BAD_RACE;
    }
    return NO_RACE;
#endif
}

void raceFound(int stmt) {
    int ot = thrille_test_mode == THRILLE_TEST_SINGLE ? 0 : (2-stmt);
    fprintf(stderr, "REAL RACE CONFIRMED (p=%lf) - Racing pairs:\n", lprob);
    fprintf(stderr, "\t(T%d:0x%"PRIx64"-%"PRIx64":%s:%s)\n", race_event.thread,
            race_event.mem.start, race_event.mem.end,
            target[stmt-1].access == READ ? "R" : "W", target[stmt-1].sig);
    fprintf(stderr, "\t(T%d:0x%"PRIx64"-%"PRIx64":%s:%s)\n",
            my_ot_race[ot].thread, my_ot_race[ot].mem.start,
            my_ot_race[ot].mem.end, 
            target[ot].access == READ ? "R" : "W", target[ot].sig);
}

#if 0
void registerWaiting() {
    upc_memput(pending + MYTHREAD, &raceEvent, sizeof(race_info_t));
    //races[MYTHREAD] = event;
}

void unregisterWaiting() {
    my_pending->addr[0] = 0;
    //races[MYTHREAD].addr[1] = 0;
}
#endif

void printRaceInfoT(race_info_t event) {
    fprintf(stderr, "race_info_t(");
    fprintf(stderr, "Thread: %d, ", event.thread);
    fprintf(stderr, "Addr: 0x%"PRIx64"-%"PRIx64", ", event.addr[0], event.addr[1]);
    fprintf(stderr, "Access Type: %s, ", event.access == WRITE ? "W" : "R");
    fprintf(stderr, "Sig: %s, ", event.sig);
    fprintf(stderr, ")\n");
}

int randTest() {
    return rand() < (lprob * RAND_MAX);
}

void randomActive(int myid, int owner, int stmt) {
    double lgprob;
    upcr_lock(locks[owner]);
    lgprob = gprob;
    lprob = lgprob < lprob ? lgprob : lprob;
    if (lprob > EPS) {
    int racingThread = checkPending(owner, stmt);
    if (racingThread == NO_RACE) {
        if (randTest()) {
            //registerWaiting();
            upc_memput(&pending[2*owner + stmt - 1], &race_event, sizeof(tester_info_t));
            semaphores[myid] = bupc_sem_alloc(BUPC_SEM_BOOLEAN | BUPC_SEM_MPRODUCER
                        | BUPC_SEM_SCONSUMER);
            int n = 0;
            fprintf(stderr, "[%d] @p%d paused at %s trying to create a real race (Addr: 0x%"PRIx64"-%"PRIx64"%s) lprob=%g\n",
                    myid, thrille_info.phase, target[stmt-1].sig,
                    race_event.mem.start, race_event.mem.end,
                    target[stmt-1].access == WRITE ? "W" : "R", lprob);
            int realRace = 0;
            bupc_tick_t tick_until = bupc_ticks_now() + timeout;
            while (bupc_ticks_now() < tick_until && 
                    !(realRace = bupc_sem_try(semaphores[myid]))) {
                upcr_unlock(locks[owner]);
                usleep(SEM_WAIT_USEC);
		bupc_poll();
                upcr_lock(locks[owner]);
            }
            //unregisterWaiting();
            pending[2*owner + stmt - 1].mem.start = 0;
            bupc_sem_free(semaphores[myid]);
            if (realRace) {
                //++num_reproduced;
                //fprintf(stderr, "%d reproduced (%d thrashes)\n", num_reproduced, num_thrash);
                if ((race_resolve_dir && stmt == 1) ||
                        (!race_resolve_dir && stmt == 2)) {
                    upcr_unlock(locks[owner]);
                    //fprintf(stderr, "[%d] Waiting guy sleeping...(dir = %d, target = %d)\n", MYTHREAD, race_resolve_dir, event.race_reported);
                    usleep(10 * SEM_WAIT_USEC);
                    upcr_lock(locks[owner]);
                }
                lprob = 0.0;
            } else {
                fprintf(stderr, "[%d] @p%d is no longer paused\n", myid, thrille_info.phase);
                //++num_thrash;
                //fprintf(stderr, "%d thrashes (%d reproduced)\n", num_thrash, num_reproduced);
                lprob *= BACK_OFF;
		gprob = lprob;
            }
        }
    } else if (racingThread != BAD_RACE) {
        raceFound(stmt);
        lprob = gprob = 0.0;
        bupc_sem_post(semaphores[racingThread]);
        if ((race_resolve_dir && stmt == 1) ||
                (!race_resolve_dir && stmt == 2)) {
            upcr_unlock(locks[owner]);
            //fprintf(stderr, "[%d] Posting guy sleeping...(dir = %d, target = %d)\n", MYTHREAD, race_resolve_dir, event.race_reported);
            usleep(10 * SEM_WAIT_USEC);
            upcr_lock(locks[owner]);
        }
    }
    }
    upcr_unlock(locks[owner]);
}

void thrille_before_read_tester(int myid,
        uint64_t mem_addr,
        const char* sig,
        int nBytes,
        int owner) {
    THRILLE_BEGIN_FUNC();

#ifdef THRILLE_DEBUG
    fprintf(stderr, "[%d] Before Read of Memory 0x%"PRIx64"-0x%"PRIx64" at %s\n", myid, mem_addr, mem_addr + nBytes, sig);
#endif

    int r;
    int owned = (MYTHREAD == owner) ? 1 : 0;
    if ((lprob > EPS) && (r = isRacingPair(sig, owned, READ))) {
        race_event.mem.start = mem_addr;
        race_event.mem.end = mem_addr + nBytes;
        race_event.phase = thrille_info.phase;
        race_event.thread = myid;
        randomActive(myid, owner, r);
#if 0
        raceEvent.addr[0] = mem_addr;
        raceEvent.addr[1] = mem_addr + nBytes;
        raceEvent.thread = myid;
        raceEvent.access = READ;
        raceEvent.sig = sig;
        //ls_strcpy(file, raceEvent.file);
        raceEvent.phase = thrille_info.phase;
        //ls_strcpy(expr, raceEvent.expr);
        raceEvent.race_reported = r;
        raceEvent.owned = owned;
        randomActive(myid, owner);
#endif
    }
    THRILLE_END_FUNC;
}

void thrille_after_read_tester(int myid,
        uint64_t mem_addr,
        const char* sig,
        int nBytes,
        int owner) {
}

uintmax_t thrille_after_read_reg_tester(int myid,
        uint64_t mem_addr,
        const char* sig,
        uintmax_t retval,
        int nBytes,
        int owner) {
    thrille_after_read_tester(myid, mem_addr, sig, nBytes, owner);
    return retval;
}

float thrille_after_read_float_tester(int myid,
        uint64_t mem_addr,
        const char* sig,
        float retval,
        int nBytes,
        int owner) {
    thrille_after_read_tester(myid, mem_addr, sig, nBytes, owner);
    return retval;
}

double thrille_after_read_double_tester(int myid,
        uint64_t mem_addr,
        const char* sig,
        double retval,
        int nBytes,
        int owner) {
    thrille_after_read_tester(myid, mem_addr, sig, nBytes, owner);
    return retval;
}

void thrille_before_write_tester(int myid,
        uint64_t mem_addr,
        const char* sig,
        int nBytes,
        int owner) {
    THRILLE_BEGIN_FUNC();

#ifdef THRILLE_DEBUG
    fprintf(stderr, "[%d] Before Write of Memory 0x%"PRIx64"-0x%"PRIx64" at %s\n", myid, mem_addr, mem_addr + nBytes, sig);
#endif

    int r;
    int owned = (MYTHREAD == owner) ? 1 : 0;
    if ((lprob > EPS) && (r = isRacingPair(sig, owned, WRITE))) {
        race_event.mem.start = mem_addr;
        race_event.mem.end = mem_addr + nBytes;
        race_event.phase = thrille_info.phase;
        race_event.thread = myid;
        randomActive(myid, owner, r);
#if 0
        raceEvent.thread = myid;
        raceEvent.addr[0] = mem_addr;
        raceEvent.addr[1] = mem_addr + nBytes;
        raceEvent.access = WRITE;
        raceEvent.sig = sig;
        //ls_strcpy(file, raceEvent.file);
        raceEvent.phase = thrille_info.phase;
        //ls_strcpy(expr, raceEvent.expr);
        raceEvent.race_reported = r;
        raceEvent.owned = owned;
        randomActive(myid, owner);
#endif
    }
    THRILLE_END_FUNC;
}

void thrille_after_write_tester(int myid,
        uint64_t mem_addr,
        const char* sig,
        int nBytes,
        int owner) {
}

void thrille_before_lread_tester(int myid,
		uint64_t mem_addr,
		const char* sig,
		int nBytes) {
	thrille_before_read_tester(myid, mem_addr, sig, nBytes, myid);
}

void thrille_after_lread_tester(int myid,
		uint64_t mem_addr,
		const char* sig,
		int nBytes) {
}

void thrille_before_lwrite_tester(int myid,
		uint64_t mem_addr,
		const char* sig,
		int nBytes) {
	thrille_before_write_tester(myid, mem_addr, sig, nBytes, myid);
}

void thrille_after_lwrite_tester(int myid,
        uint64_t mem_addr,
        const char* sig,
        int nBytes) {
}

void thrille_before_lock_tester(int myid,
        uint64_t lock,
        const char* sig) {
}

void thrille_after_lock_tester(int myid,
        uint64_t lock_addr,
        const char* sig) {
    THRILLE_BEGIN_FUNC();
    //fprintf(stderr, "[%d] After lock 0x%lx\n", myid, lock_addr);
    THRILLE_END_FUNC;
}

void thrille_before_lock_attempt_tester(int myid,
        uint64_t lock,
        const char* sig) {
}

int thrille_after_lock_attempt_tester(int myid,
        uint64_t lock,
        const char* sig,
        int retval) {

#ifdef THRILLE_DEBUG
    fprintf(stderr, "[%d] After lock_attempt 0x%"PRIx64" (%s)\n", myid, lock, retval ? "success" : "fail");
#endif

    if (retval) {
        thrille_after_lock_tester(myid, lock, sig);
    }

    return retval;
}

void thrille_before_unlock_tester(int myid,
        uint64_t lock_addr,
        const char* sig) {
    THRILLE_BEGIN_FUNC();
    //fprintf(stderr, "[%d] Before unlock 0x%lx\n", myid, lock_addr);
    THRILLE_END_FUNC;
}

void thrille_after_unlock_tester(int myid,
        uint64_t lock,
        const char* sig) {
}

void thrille_before_notify_tester(int myid, int barrier_val, const char* sig) {
#if 0
    THRILLE_BEGIN_FUNC();
    if (lprob > EPS && (myid == ((thrille_info.phase/2) % THREADS))) {
        gprob *= lprob;
    }
    THRILLE_END_FUNC;
#endif
}

void thrille_after_notify_tester(int myid, int barrier_val, const char* sig) {
    THRILLE_BEGIN_FUNC();
    thrille_info.phase++;
    //fprintf(stderr, "[%d] Now in phase %d (After notify)\n", myid, thrille_info.phase);
    THRILLE_END_FUNC;
}

void thrille_before_wait_tester(int myid, int barrier_val, const char* sig) {
}

void thrille_after_wait_tester(int myid, int barrier_val, const char* sig) {
    THRILLE_BEGIN_FUNC();
    thrille_info.phase++;
    //fprintf(stderr, "[%d] Now in phase %d (After wait)\n", myid, thrille_info.phase);
#if 0
    if (lprob > EPS) {
        lprob *= gprob;
    }
    /*if (myid == 0)
        fprintf(stderr, "lprob[0] = %g\n", lprob);*/
#endif
    THRILLE_END_FUNC;
}

void thrille_before_try_wait_tester(int myid, int barrier_val, const char* sig) {
}

int thrille_after_try_wait_tester(int myid, int barrier_val, const char* sig, int retval) {
    if (retval) {
        thrille_after_wait_tester(myid, barrier_val, sig);
    }
    return retval;
}

#include "upct_common.c"
